@@ -102,8 +102,8 @@ QT_BEGIN_NAMESPACE
     these settings can also be controlled interactively when a \l DebugView
     item is added to the scene.
 
-    \li Fog settings. To enable fog, set an appropriately configured \l Fog
-    object in the \l fog property.
+    \li Fog settings. To enable fog, set an appropriately configured
+    \l [QML] Fog object in the \l [QML] fog property.
 
     \li Lightmap baking settings. When pre-baked lightmaps are used for some
     models in the scene, the \l Lightmapper object set in the \l lightmapper
@@ -452,7 +452,8 @@ float QQuick3DSceneEnvironment::aoDistance() const
 /*!
     \qmlproperty float QtQuick3D::SceneEnvironment::aoSoftness
 
-    This property how smooth the edges of the ambient occlusion shading are.
+    This property defines how smooth the edges of the ambient occlusion shading
+    are.
 
     The value must be between 0.0 and 50.0. The default value is 50.0.
 
@@ -1463,7 +1464,7 @@ void QQuick3DSceneEnvironment::setGridFlags(uint newGridFlags)
 }
 
 /*!
-    \qmlproperty float SceneEnvironment::aoEnabled
+    \qmlproperty bool SceneEnvironment::aoEnabled
     \since 6.5
 
     Enable or disable ambient occlusion.
