@@ -33,6 +33,7 @@
 #include <QtGui/private/qeventpoint_p.h>
 
 #include <QtCore/private/qnumeric_p.h>
+#include <QtCore/qpointer.h>
 
 #include <optional>
 
@@ -70,7 +71,15 @@ struct ViewportTransformHelper : public QQuickDeliveryAgent::Transform
         If it's no longer a "hit" on sceneParentNode, returns the last-good point.
     */
     QPointF map(const QPointF &viewportPoint) override {
-        std::optional<QSSGRenderRay> rayResult = renderer->getRayFromViewportPos(viewportPoint * dpr);
+        QPointF point = viewportPoint;
+        // Despite the name, the input coordinates are the window viewport coordinates
+        // so unless the View3D is the same size of the Window, we need to translate
+        // to the View3D coordinates before doing any picking.
+        if (viewport)
+            point = viewport->mapFromScene(viewportPoint);
+        point.rx() *= scaleX;
+        point.ry() *= scaleY;
+        std::optional<QSSGRenderRay> rayResult = renderer->getRayFromViewportPos(point);
         if (rayResult.has_value()) {
             auto pickResult = renderer->syncPickOne(rayResult.value(), sceneParentNode);
             auto ret = pickResult.m_localUVCoords.toPointF();
@@ -91,10 +100,12 @@ struct ViewportTransformHelper : public QQuickDeliveryAgent::Transform
         return QPointF();
     }
 
+    QPointer<QQuick3DViewport> viewport;
     QQuick3DSceneRenderer *renderer = nullptr;
     QSSGRenderNode *sceneParentNode = nullptr;
-    QQuickItem* targetItem = nullptr;
-    qreal dpr = 1;
+    QPointer<QQuickItem> targetItem;
+    qreal scaleX = 1;
+    qreal scaleY = 1;
     bool uvCoordsArePixels = false; // if false, they are in the range 0..1
     QPointF lastGoodMapping;
 
@@ -252,6 +263,18 @@ QQuick3DViewport::QQuick3DViewport(QQuickItem *parent)
 
 QQuick3DViewport::~QQuick3DViewport()
 {
+    // With the threaded render loop m_directRenderer must be destroyed on the
+    // render thread at the proper time, not here. That's handled in
+    // releaseResources() + upon sceneGraphInvalidated. However with a
+    // QQuickRenderControl-based window on the main thread there is a good
+    // chance that this viewport (and so our sceneGraphInvalidated signal
+    // connection) is destroyed before the window and the rendercontrol. So act
+    // here then.
+    if (m_directRenderer && m_directRenderer->thread() == thread()) {
+        delete m_directRenderer;
+        m_directRenderer = nullptr;
+    }
+
     // If the quick window still exists, make sure to disconnect any of the direct
     // connections to this View3D
     if (auto qw = window())
@@ -271,13 +294,8 @@ QQuick3DViewport::~QQuick3DViewport()
     // might still be rendering.
     m_renderStats->deleteLater();
 
-    if (!window() && sceneManager && sceneManager->wattached) {
-        if (sceneManager->wattached->rci().use_count() <= 1)
-            delete sceneManager->wattached;
-    }
-
-    // m_directRenderer must be destroyed on the render thread at the proper time, not here.
-    // That's handled in releaseResources() + upon sceneGraphInvalidated
+    if (!window() && sceneManager && sceneManager->wattached)
+        QMetaObject::invokeMethod(sceneManager->wattached, &QQuick3DWindowAttachment::evaluateEol, Qt::QueuedConnection);
 }
 
 static void ssgn_append(QQmlListProperty<QObject> *property, QObject *obj)
@@ -768,6 +786,94 @@ void QQuick3DViewport::setRenderFormat(QQuickShaderEffectSource::Format format)
     update();
 }
 
+/*!
+    \qmlproperty int QtQuick3D::View3D::explicitTextureWidth
+    \since 6.7
+
+    The width, in pixels, of the item's associated texture. Relevant when a
+    fixed texture size is desired that does not depend on the item's size. This
+    size has no effect on the geometry of the item (its size and placement
+    within the scene), which means the texture's content will appear scaled up
+    or down (and possibly stretched) onto the item's area.
+
+    By default the value is \c 0. A value of 0 means that texture's size
+    follows the item's size. (\c{texture size in pixels} = \c{item's logical
+    size} * \c{device pixel ratio}).
+
+    \note This property is relevant only when \l renderMode is set to \c
+    Offscreen. Its value is ignored otherwise.
+
+    \sa explicitTextureHeight, effectiveTextureSize, DebugView
+*/
+int QQuick3DViewport::explicitTextureWidth() const
+{
+    return m_explicitTextureWidth;
+}
+
+void QQuick3DViewport::setExplicitTextureWidth(int width)
+{
+    if (m_explicitTextureWidth == width)
+        return;
+
+    m_explicitTextureWidth = width;
+    emit explicitTextureWidthChanged();
+    update();
+}
+
+/*!
+    \qmlproperty int QtQuick3D::View3D::explicitTextureHeight
+    \since 6.7
+
+    The height, in pixels, of the item's associated texture. Relevant when a
+    fixed texture size is desired that does not depend on the item's size. This
+    size has no effect on the geometry of the item (its size and placement
+    within the scene), which means the texture's content will appear scaled up
+    or down (and possibly stretched) onto the item's area.
+
+    By default the value is \c 0. A value of 0 means that texture's size
+    follows the item's size. (\c{texture size in pixels} = \c{item's logical
+    size} * \c{device pixel ratio}).
+
+    \note This property is relevant only when \l renderMode is set to \c
+    Offscreen. Its value is ignored otherwise.
+
+    \sa explicitTextureWidth, effectiveTextureSize, DebugView
+*/
+int QQuick3DViewport::explicitTextureHeight() const
+{
+    return m_explicitTextureHeight;
+}
+
+void QQuick3DViewport::setExplicitTextureHeight(int height)
+{
+    if (m_explicitTextureHeight == height)
+        return;
+
+    m_explicitTextureHeight = height;
+    emit explicitTextureHeightChanged();
+    update();
+}
+
+/*!
+    \qmlproperty size QtQuick3D::View3D::effectiveTextureSize
+    \since 6.7
+
+    This property exposes the size, in pixels, of the underlying color (and
+    depth/stencil) buffers. It is provided for use on the GUI (main) thread, in
+    QML bindings or JavaScript.
+
+    This is a read-only property.
+
+    \note This property is relevant only when \l renderMode is set to
+    \c Offscreen.
+
+    \sa explicitTextureWidth, explicitTextureHeight, DebugView
+*/
+QSize QQuick3DViewport::effectiveTextureSize() const
+{
+    return m_effectiveTextureSize;
+}
+
 
 /*!
     \qmlmethod vector3d View3D::mapFrom3DScene(vector3d scenePos)
@@ -848,8 +954,8 @@ QQuick3DPickResult QQuick3DViewport::pick(float x, float y) const
     if (!renderer)
         return QQuick3DPickResult();
 
-    const QPointF position(qreal(x) * window()->effectiveDevicePixelRatio(),
-                           qreal(y) * window()->effectiveDevicePixelRatio());
+    const QPointF position(qreal(x) * window()->effectiveDevicePixelRatio() * m_widthMultiplier,
+                           qreal(y) * window()->effectiveDevicePixelRatio() * m_heightMultiplier);
     std::optional<QSSGRenderRay> rayResult = renderer->getRayFromViewportPos(position);
     if (!rayResult.has_value())
         return QQuick3DPickResult();
@@ -876,8 +982,8 @@ QList<QQuick3DPickResult> QQuick3DViewport::pickAll(float x, float y) const
     if (!renderer)
         return QList<QQuick3DPickResult>();
 
-    const QPointF position(qreal(x) * window()->effectiveDevicePixelRatio(),
-                           qreal(y) * window()->effectiveDevicePixelRatio());
+    const QPointF position(qreal(x) * window()->effectiveDevicePixelRatio() * m_widthMultiplier,
+                           qreal(y) * window()->effectiveDevicePixelRatio() * m_heightMultiplier);
     std::optional<QSSGRenderRay> rayResult = renderer->getRayFromViewportPos(position);
     if (!rayResult.has_value())
         return QList<QQuick3DPickResult>();
@@ -1043,12 +1149,30 @@ QSGNode *QQuick3DViewport::setupOffscreenRenderer(QSGNode *node)
         n->quickFbo = this;
         connect(window(), SIGNAL(screenChanged(QScreen*)), n, SLOT(handleScreenChange()));
     }
-    QSize minFboSize = QQuickItemPrivate::get(this)->sceneGraphContext()->minimumFBOSize();
-    QSize desiredFboSize(qMax<int>(minFboSize.width(), width()),
-                         qMax<int>(minFboSize.height(), height()));
 
-    n->devicePixelRatio = window()->effectiveDevicePixelRatio();
-    desiredFboSize *= n->devicePixelRatio;
+    const qreal dpr = window()->effectiveDevicePixelRatio();
+    const QSize minFboSize = QQuickItemPrivate::get(this)->sceneGraphContext()->minimumFBOSize();
+    QSize desiredFboSize = QSize(m_explicitTextureWidth, m_explicitTextureHeight);
+    if (desiredFboSize.isEmpty()) {
+        desiredFboSize = QSize(width(), height()) * dpr;
+        n->devicePixelRatio = dpr;
+        // 1:1 mapping between the backing texture and the on-screen quad
+        m_widthMultiplier = 1.0f;
+        m_heightMultiplier = 1.0f;
+    } else {
+        QSize itemPixelSize = QSize(width(), height()) * dpr;
+        // not 1:1 maping between the backing texture and the on-screen quad
+        m_widthMultiplier = desiredFboSize.width() / float(itemPixelSize.width());
+        m_heightMultiplier = desiredFboSize.height() / float(itemPixelSize.height());
+        n->devicePixelRatio = 1.0;
+    }
+    desiredFboSize.setWidth(qMax(minFboSize.width(), desiredFboSize.width()));
+    desiredFboSize.setHeight(qMax(minFboSize.height(), desiredFboSize.height()));
+
+    if (desiredFboSize != m_effectiveTextureSize) {
+        m_effectiveTextureSize = desiredFboSize;
+        emit effectiveTextureSizeChanged();
+    }
 
     n->setFiltering(smooth() ? QSGTexture::Linear : QSGTexture::Nearest);
     n->setRect(0, 0, width(), height());
@@ -1079,6 +1203,11 @@ QSGNode *QQuick3DViewport::setupInlineRenderer(QSGNode *node)
             return nullptr;
     }
 
+    if (!m_effectiveTextureSize.isEmpty()) {
+        m_effectiveTextureSize = QSize();
+        emit effectiveTextureSizeChanged();
+    }
+
     const QSize targetSize = window()->effectiveDevicePixelRatio() * QSize(width(), height());
 
     // checkIsVisible, not isVisible, because, for example, a
@@ -1106,6 +1235,11 @@ void QQuick3DViewport::setupDirectRenderer(RenderMode mode)
         connect(window(), &QQuickWindow::sceneGraphInvalidated, this, &QQuick3DViewport::cleanupDirectRenderer, Qt::DirectConnection);
     }
 
+    if (!m_effectiveTextureSize.isEmpty()) {
+        m_effectiveTextureSize = QSize();
+        emit effectiveTextureSizeChanged();
+    }
+
     const QSizeF targetSize = window()->effectiveDevicePixelRatio() * QSizeF(width(), height());
     m_directRenderer->setViewport(QRectF(window()->effectiveDevicePixelRatio() * mapToScene(QPointF(0, 0)), targetSize));
     m_directRenderer->setVisibility(isVisible());
@@ -1160,12 +1294,12 @@ void QQuick3DViewport::processPickedObject(const QSSGRenderGraphObject *backendO
         if (!subsceneRootItem || subsceneRootItem->childItems().isEmpty())
             return; // ignore empty 2D subscenes
 
-        // In this case the "UV" coordinates are in pixels in the subscene root item, so we can just use them.
+        // In this case the "UV" coordinates are in pixels in the subscene root item's coordinate system.
         subscenePosition = pickResult.m_localUVCoords.toPointF();
-        // Even though an Item2D is an "infinite plane" for rendering purposes,
-        // avoid delivering events outside the rectangular area that is occupied by child items,
-        // so that events can "fall through" to other interactive content in the scene or behind it.
-        if (!subsceneRootItem->childrenRect().contains(subscenePosition))
+
+        // The following code will account for custom input masking, as well any
+        // transformations that might have been applied to the Item
+        if (!subsceneRootItem->childAt(subscenePosition.x(), subscenePosition.y()))
             return;
     } else if (frontendObjectPrivate->type == QQuick3DObjectPrivate::Type::Model) {
         // Model
@@ -1310,10 +1444,12 @@ bool QQuick3DViewport::forwardEventToSubscenes(QPointerEvent *event,
                 auto frontendObjectPrivate = QQuick3DObjectPrivate::get(subsceneInfo.obj);
                 const bool item2Dcase = (frontendObjectPrivate->type == QQuick3DObjectPrivate::Type::Item2D);
                 ViewportTransformHelper *transform = new ViewportTransformHelper;
+                transform->viewport = const_cast<QQuick3DViewport *>(this);
                 transform->renderer = renderer;
                 transform->sceneParentNode = static_cast<QSSGRenderNode*>(frontendObjectPrivate->spatialNode);
                 transform->targetItem = subsceneRoot;
-                transform->dpr = window()->effectiveDevicePixelRatio();
+                transform->scaleX = window()->effectiveDevicePixelRatio() * m_widthMultiplier;
+                transform->scaleY = window()->effectiveDevicePixelRatio() * m_heightMultiplier;
                 transform->uvCoordsArePixels = item2Dcase;
                 transform->setOnDeliveryAgent(da);
                 qCDebug(lcPick) << event->type() << "created ViewportTransformHelper on" << da;
@@ -1371,7 +1507,10 @@ QVarLengthArray<QSSGRenderPickResult, 20> QQuick3DViewport::getPickResults(QQuic
 QVarLengthArray<QSSGRenderPickResult, 20> QQuick3DViewport::getPickResults(QQuick3DSceneRenderer *renderer, const QEventPoint &eventPoint) const
 {
     QVarLengthArray<QSSGRenderPickResult, 20> pickResults;
-    const QPointF realPosition = eventPoint.position() * window()->effectiveDevicePixelRatio();
+    QPointF realPosition = eventPoint.position() * window()->effectiveDevicePixelRatio();
+    // correct when mapping is not 1:1 due to explicit backing texture size
+    realPosition.rx() *= m_widthMultiplier;
+    realPosition.ry() *= m_heightMultiplier;
     std::optional<QSSGRenderRay> rayResult = renderer->getRayFromViewportPos(realPosition);
     if (rayResult.has_value())
         pickResults = renderer->syncPickAll(rayResult.value());
@@ -1450,7 +1589,11 @@ void QQuick3DViewport::onReleaseCachedResources()
 }
 
 /*!
-    \internal
+    \qmlproperty List<QtQuick3D::Object3D> View3D::extensions
+
+    This property contains a list of user extensions that should be used with this \l View3D.
+
+    \sa RenderExtension
 */
 QQmlListProperty<QQuick3DObject> QQuick3DViewport::extensions()
 {
