@@ -6,7 +6,7 @@
 
 #include <QtQuick3DUtils/private/qssgutils_p.h>
 
-#include <QtQuick3DRuntimeRender/private/qssgrendercontextcore_p.h>
+#include "qssgrendercontextcore.h"
 #include <QtQuick3DRuntimeRender/private/qssgrenderbuffermanager_p.h>
 #include <QtQuick3DRuntimeRender/private/qssgrendermesh_p.h>
 #include <QtQuick3DRuntimeRender/private/qssgrendercamera_p.h>
@@ -53,6 +53,7 @@ void QSSGCustomMaterialSystem::releaseCachedResources()
 QSSGRhiShaderPipelinePtr QSSGCustomMaterialSystem::shadersForCustomMaterial(QSSGRhiGraphicsPipelineState *ps,
                                                                             const QSSGRenderCustomMaterial &material,
                                                                             QSSGSubsetRenderable &renderable,
+                                                                            const QSSGShaderDefaultMaterialKeyProperties &defaultMaterialShaderKeyProperties,
                                                                             const QSSGShaderFeatures &featureSet)
 {
     QElapsedTimer timer;
@@ -73,7 +74,7 @@ QSSGRhiShaderPipelinePtr QSSGCustomMaterialSystem::shadersForCustomMaterial(QSSG
         // NB this key calculation must replicate exactly what the generator does in generateMaterialRhiShader()
         QByteArray shaderString = material.m_shaderPathKey;
         QSSGShaderDefaultMaterialKey matKey(renderable.shaderDescription);
-        matKey.toString(shaderString, context->renderer()->defaultMaterialShaderKeyProperties());
+        matKey.toString(shaderString, defaultMaterialShaderKeyProperties);
 
         // Try the persistent (disk-based) cache.
         const QByteArray qsbcKey = QQsbCollection::EntryDesc::generateSha(shaderString, QQsbCollection::toFeatureSet(featureSet));
@@ -84,13 +85,13 @@ QSSGRhiShaderPipelinePtr QSSGCustomMaterialSystem::shadersForCustomMaterial(QSSG
             Q_TRACE_SCOPE(QSSG_generateShader);
             Q_QUICK3D_PROFILE_START(QQuick3DProfiler::Quick3DGenerateShader);
             QSSGMaterialVertexPipeline vertexPipeline(*context->shaderProgramGenerator(),
-                                                      context->renderer()->defaultMaterialShaderKeyProperties(),
+                                                      defaultMaterialShaderKeyProperties,
                                                       material.adapter);
 
             shaderPipeline = QSSGMaterialShaderGenerator::generateMaterialRhiShader(material.m_shaderPathKey,
                                                                                     vertexPipeline,
                                                                                     renderable.shaderDescription,
-                                                                                    context->renderer()->defaultMaterialShaderKeyProperties(),
+                                                                                    defaultMaterialShaderKeyProperties,
                                                                                     featureSet,
                                                                                     renderable.material,
                                                                                     renderable.lights,
@@ -100,7 +101,7 @@ QSSGRhiShaderPipelinePtr QSSGCustomMaterialSystem::shadersForCustomMaterial(QSSG
             Q_QUICK3D_PROFILE_END_WITH_ID(QQuick3DProfiler::Quick3DGenerateShader, 0, material.profilingId);
         }
 
-        // make skey useable as a key for the QHash (makes copies of materialKey and featureSet, instead of just referencing)
+        // make skey useable as a key for the QHash (makes a copy of the materialKey, instead of just referencing)
         skey.detach();
         // insert it no matter what, no point in trying over and over again
         shaderMap.insert(skey, shaderPipeline);
@@ -109,17 +110,18 @@ QSSGRhiShaderPipelinePtr QSSGCustomMaterialSystem::shadersForCustomMaterial(QSSG
     }
 
     if (shaderPipeline) {
-        ps->shaderPipeline = shaderPipeline.get();
+        QSSGRhiGraphicsPipelineStatePrivate::setShaderPipeline(*ps, shaderPipeline.get());
         shaderPipeline->resetExtraTextures();
     }
 
-    context->rhiContext()->stats().registerMaterialShaderGenerationTime(timer.elapsed());
+    QSSGRhiContextStats::get(*context->rhiContext()).registerMaterialShaderGenerationTime(timer.elapsed());
 
     return shaderPipeline;
 }
 
 void QSSGCustomMaterialSystem::updateUniformsForCustomMaterial(QSSGRhiShaderPipeline &shaderPipeline,
                                                                QSSGRhiContext *rhiCtx,
+                                                               const QSSGLayerRenderData &inData,
                                                                char *ubufData,
                                                                QSSGRhiGraphicsPipelineState *ps,
                                                                const QSSGRenderCustomMaterial &material,
@@ -132,11 +134,15 @@ void QSSGCustomMaterialSystem::updateUniformsForCustomMaterial(QSSGRhiShaderPipe
                                                      : renderable.modelContext.modelViewProjection);
 
     const QMatrix4x4 clipSpaceCorrMatrix = rhiCtx->rhi()->clipSpaceCorrMatrix();
+    QRhiTexture *lightmapTexture = inData.getLightmapTexture(renderable.modelContext);
 
     const auto &modelNode = renderable.modelContext.model;
     const QMatrix4x4 &localInstanceTransform(modelNode.localInstanceTransform);
     const QMatrix4x4 &globalInstanceTransform(modelNode.globalInstanceTransform);
-    const QMatrix4x4 &modelMatrix((modelNode.boneCount == 0) ? renderable.globalTransform : QMatrix4x4());
+
+    const auto &defaultMaterialShaderKeyProperties = inData.getDefaultMaterialPropertyTable();
+
+    const QMatrix4x4 &modelMatrix(modelNode.usesBoneTexture() ? QMatrix4x4() : renderable.globalTransform);
 
     QSSGMaterialShaderGenerator::setRhiMaterialProperties(*context,
                                                           shaderPipeline,
@@ -144,7 +150,7 @@ void QSSGCustomMaterialSystem::updateUniformsForCustomMaterial(QSSGRhiShaderPipe
                                                           ps,
                                                           material,
                                                           renderable.shaderDescription,
-                                                          context->renderer()->defaultMaterialShaderKeyProperties(),
+                                                          defaultMaterialShaderKeyProperties,
                                                           camera,
                                                           mvp,
                                                           renderable.modelContext.normalMatrix,
@@ -155,13 +161,13 @@ void QSSGCustomMaterialSystem::updateUniformsForCustomMaterial(QSSGRhiShaderPipe
                                                           toDataView(modelNode.morphWeights),
                                                           renderable.firstImage,
                                                           renderable.opacity,
-                                                          renderable.renderer->getLayerGlobalRenderProperties(),
+                                                          inData,
                                                           renderable.lights,
                                                           renderable.reflectionProbe,
                                                           true,
                                                           renderable.renderableFlags.receivesReflections(),
                                                           depthAdjust,
-                                                          renderable.modelContext.lightmapTexture);
+                                                          lightmapTexture);
 }
 
 static const QRhiShaderResourceBinding::StageFlags CUSTOM_MATERIAL_VISIBILITY_ALL =
@@ -176,7 +182,7 @@ void QSSGCustomMaterialSystem::rhiPrepareRenderable(QSSGRhiGraphicsPipelineState
                                                     QRhiRenderPassDescriptor *renderPassDescriptor,
                                                     int samples,
                                                     QSSGRenderCamera *camera,
-                                                    int cubeFace,
+                                                    QSSGRenderTextureCubeFace cubeFace,
                                                     QMatrix4x4 *modelViewProjection,
                                                     QSSGReflectionMapEntry *entry)
 {
@@ -186,37 +192,42 @@ void QSSGCustomMaterialSystem::rhiPrepareRenderable(QSSGRhiGraphicsPipelineState
     if (material.m_renderFlags.testFlag(QSSGRenderCustomMaterial::RenderFlag::Blending)) {
         blend.enable = true;
         blend.srcColor = material.m_srcBlend;
-        blend.srcAlpha = material.m_srcBlend;
+        blend.srcAlpha = material.m_srcAlphaBlend;
         blend.dstColor = material.m_dstBlend;
-        blend.dstAlpha = material.m_dstBlend;
+        blend.dstAlpha = material.m_dstAlphaBlend;
     }
 
     const QSSGCullFaceMode cullMode = material.m_cullMode;
 
-    const bool blendParticles = renderable.renderer->defaultMaterialShaderKeyProperties().m_blendParticles.getValue(renderable.shaderDescription);
+    const auto &defaultMaterialShaderKeyProperties = layerData.getDefaultMaterialPropertyTable();
 
-    const auto &shaderPipeline = shadersForCustomMaterial(ps, material, renderable, featureSet);
+    const bool blendParticles = defaultMaterialShaderKeyProperties.m_blendParticles.getValue(renderable.shaderDescription);
+
+    const auto &shaderPipeline = shadersForCustomMaterial(ps, material, renderable, defaultMaterialShaderKeyProperties, featureSet);
 
     if (shaderPipeline) {
         QSSGRhiShaderResourceBindingList bindings;
         const auto &modelNode = renderable.modelContext.model;
 
-        QSSGRhiDrawCallData &dcd(cubeFace < 0 ? rhiCtx->drawCallData({ passKey,
-                                                        &modelNode,
-                                                        &material,
-                                                        0,
-                                                        QSSGRhiDrawCallDataKey::Main })
-                                              : rhiCtx->drawCallData({ passKey,
-                                                                       &modelNode,
-                                                                       entry, cubeFace + int(renderable.subset.offset << 3),
-                                                                       QSSGRhiDrawCallDataKey::Reflection }));
+        // NOTE:
+        // - entryIdx should 0 for QSSGRenderTextureCubeFaceNone.
+        // In all other cases the entryIdx is a combination of the cubeface idx and the subset offset, where the lower bits
+        // are the cubeface idx.
+        const auto cubeFaceIdx = QSSGBaseTypeHelpers::indexOfCubeFace(cubeFace);
+        const quintptr entryIdx = quintptr(cubeFace != QSSGRenderTextureCubeFaceNone) * (cubeFaceIdx + (quintptr(renderable.subset.offset) << 3));
+        // As the entry might be null we create an entry key consisting of the entry and the material.
+        const auto entryPartA = reinterpret_cast<quintptr>(&material);
+        const auto entryPartB = reinterpret_cast<quintptr>(entry);
+        const void *entryKey = reinterpret_cast<const void *>(entryPartA ^ entryPartB);
+
+        QSSGRhiDrawCallData &dcd = QSSGRhiContextPrivate::get(rhiCtx)->drawCallData({ passKey, &modelNode, entryKey, entryIdx });
 
         shaderPipeline->ensureCombinedMainLightsUniformBuffer(&dcd.ubuf);
         char *ubufData = dcd.ubuf->beginFullDynamicBufferUpdateForCurrentFrame();
         if (!camera)
-            updateUniformsForCustomMaterial(*shaderPipeline, rhiCtx, ubufData, ps, material, renderable, *layerData.camera, nullptr, nullptr);
+            updateUniformsForCustomMaterial(*shaderPipeline, rhiCtx, layerData, ubufData, ps, material, renderable, *layerData.camera, nullptr, nullptr);
         else
-            updateUniformsForCustomMaterial(*shaderPipeline, rhiCtx, ubufData, ps, material, renderable, *camera, nullptr, modelViewProjection);
+            updateUniformsForCustomMaterial(*shaderPipeline, rhiCtx, layerData, ubufData, ps, material, renderable, *camera, nullptr, modelViewProjection);
         if (blendParticles)
             QSSGParticleRenderer::updateUniformsForParticleModel(*shaderPipeline, ubufData, &renderable.modelContext.model, renderable.subset.offset);
         dcd.ubuf->endFullDynamicBufferUpdateForCurrentFrame();
@@ -225,17 +236,19 @@ void QSSGCustomMaterialSystem::rhiPrepareRenderable(QSSGRhiGraphicsPipelineState
             QSSGParticleRenderer::prepareParticlesForModel(*shaderPipeline, rhiCtx, bindings, &renderable.modelContext.model);
         bool instancing = false;
         if (!camera)
-            instancing = renderable.prepareInstancing(rhiCtx, layerData.cameraData->direction, layerData.cameraData->position, renderable.instancingLodMin, renderable.instancingLodMax);
+            instancing = QSSGLayerRenderData::prepareInstancing(rhiCtx, &renderable, layerData.cameraData->direction, layerData.cameraData->position, renderable.instancingLodMin, renderable.instancingLodMax);
         else
-            instancing = renderable.prepareInstancing(rhiCtx, camera->getScalingCorrectDirection(), camera->getGlobalPos(), renderable.instancingLodMin, renderable.instancingLodMax);
+            instancing = QSSGLayerRenderData::prepareInstancing(rhiCtx, &renderable, camera->getScalingCorrectDirection(), camera->getGlobalPos(), renderable.instancingLodMin, renderable.instancingLodMax);
 
         ps->samples = samples;
 
-        ps->cullMode = QSSGRhiGraphicsPipelineState::toCullMode(cullMode);
+        ps->cullMode = QSSGRhiHelpers::toCullMode(cullMode);
 
         ps->targetBlend = blend;
 
-        ps->ia = renderable.subset.rhi.ia;
+        auto &ia = QSSGRhiInputAssemblerStatePrivate::get(*ps);
+
+        ia = renderable.subset.rhi.ia;
 
         //### Copied code from default materials
         int instanceBufferBinding = 0;
@@ -243,15 +256,15 @@ void QSSGCustomMaterialSystem::rhiPrepareRenderable(QSSGRhiGraphicsPipelineState
             // Need to setup new bindings for instanced buffers
             const quint32 stride = renderable.modelContext.model.instanceTable->stride();
             QVarLengthArray<QRhiVertexInputBinding, 8> bindings;
-            std::copy(ps->ia.inputLayout.cbeginBindings(),
-                      ps->ia.inputLayout.cendBindings(),
+            std::copy(ia.inputLayout.cbeginBindings(),
+                      ia.inputLayout.cendBindings(),
                       std::back_inserter(bindings));
             bindings.append({ stride, QRhiVertexInputBinding::PerInstance });
             instanceBufferBinding = bindings.size() - 1;
-            ps->ia.inputLayout.setBindings(bindings.cbegin(), bindings.cend());
+            ia.inputLayout.setBindings(bindings.cbegin(), bindings.cend());
         }
 
-        ps->ia.bakeVertexInputLocations(*shaderPipeline, instanceBufferBinding);
+        QSSGRhiHelpers::bakeVertexInputLocations(&ia, *shaderPipeline, instanceBufferBinding);
 
         QRhiResourceUpdateBatch *resourceUpdates = rhiCtx->rhi()->nextResourceUpdateBatch();
         QRhiTexture *dummyTexture = rhiCtx->dummyTexture({}, resourceUpdates);
@@ -289,12 +302,7 @@ void QSSGCustomMaterialSystem::rhiPrepareRenderable(QSSGRhiGraphicsPipelineState
             samplerBindingsSpecified.setBit(shaderPipeline->bindingForTexture("qt_particleTexture"));
 
         // Skinning
-        if (modelNode.boneCount != 0) {
-            QRhiResourceUpdateBatch *rub = rhiCtx->rhi()->nextResourceUpdateBatch();
-            QRhiTextureSubresourceUploadDescription boneDesc(modelNode.boneData);
-            QRhiTextureUploadDescription boneUploadDesc(QRhiTextureUploadEntry(0, 0, boneDesc));
-            rub->uploadTexture(modelNode.boneTexture, boneUploadDesc);
-            rhiCtx->commandBuffer()->resourceUpdate(rub);
+        if (QRhiTexture *boneTexture = layerData.getBonemapTexture(renderable.modelContext)) {
             int binding = shaderPipeline->bindingForTexture("qt_boneTexture");
             if (binding >= 0) {
                 QRhiSampler *boneSampler = rhiCtx->sampler({ QRhiSampler::Nearest,
@@ -304,7 +312,27 @@ void QSSGCustomMaterialSystem::rhiPrepareRenderable(QSSGRhiGraphicsPipelineState
                                                              QRhiSampler::ClampToEdge,
                                                              QRhiSampler::Repeat
                                                            });
-                bindings.addTexture(binding, QRhiShaderResourceBinding::VertexStage, modelNode.boneTexture, boneSampler);
+                bindings.addTexture(binding,
+                                    QRhiShaderResourceBinding::VertexStage,
+                                    boneTexture,
+                                    boneSampler);
+                samplerBindingsSpecified.setBit(binding);
+            }
+        }
+
+        // Morphing
+        auto *targetsTexture = renderable.subset.rhi.targetsTexture;
+        if (targetsTexture) {
+            int binding = shaderPipeline->bindingForTexture("qt_morphTargetTexture");
+            if (binding >= 0) {
+                QRhiSampler *targetsSampler = rhiCtx->sampler({ QRhiSampler::Nearest,
+                                                                QRhiSampler::Nearest,
+                                                                QRhiSampler::None,
+                                                                QRhiSampler::ClampToEdge,
+                                                                QRhiSampler::ClampToEdge,
+                                                                QRhiSampler::ClampToEdge
+                                                           });
+                bindings.addTexture(binding, QRhiShaderResourceBinding::VertexStage, renderable.subset.rhi.targetsTexture, targetsSampler);
                 samplerBindingsSpecified.setBit(binding);
             }
         }
@@ -327,7 +355,7 @@ void QSSGCustomMaterialSystem::rhiPrepareRenderable(QSSGRhiGraphicsPipelineState
                 samplerBindingsSpecified.setBit(binding);
                 QPair<QSSGRenderTextureCoordOp, QSSGRenderTextureCoordOp> tiling = shaderPipeline->lightProbeTiling();
                 QRhiSampler *sampler = rhiCtx->sampler({ QRhiSampler::Linear, QRhiSampler::Linear, QRhiSampler::Linear, // enables mipmapping
-                                                         toRhi(tiling.first), toRhi(tiling.second), QRhiSampler::Repeat });
+                                                         QSSGRhiHelpers::toRhi(tiling.first), QSSGRhiHelpers::toRhi(tiling.second), QRhiSampler::Repeat });
                 bindings.addTexture(binding,
                                     QRhiShaderResourceBinding::FragmentStage,
                                     shaderPipeline->lightProbeTexture(), sampler);
@@ -418,12 +446,12 @@ void QSSGCustomMaterialSystem::rhiPrepareRenderable(QSSGRhiGraphicsPipelineState
                 if (samplerBinding >= 0 && texture) {
                     const bool mipmapped = texture->flags().testFlag(QRhiTexture::MipMapped);
                     QSSGRhiSamplerDescription samplerDesc = {
-                        toRhi(renderableImage->m_imageNode.m_minFilterType),
-                        toRhi(renderableImage->m_imageNode.m_magFilterType),
-                        mipmapped ? toRhi(renderableImage->m_imageNode.m_mipFilterType) : QRhiSampler::None,
-                        toRhi(renderableImage->m_imageNode.m_horizontalTilingMode),
-                        toRhi(renderableImage->m_imageNode.m_verticalTilingMode),
-                        QRhiSampler::Repeat
+                        QSSGRhiHelpers::toRhi(renderableImage->m_imageNode.m_minFilterType),
+                        QSSGRhiHelpers::toRhi(renderableImage->m_imageNode.m_magFilterType),
+                        mipmapped ? QSSGRhiHelpers::toRhi(renderableImage->m_imageNode.m_mipFilterType) : QRhiSampler::None,
+                        QSSGRhiHelpers::toRhi(renderableImage->m_imageNode.m_horizontalTilingMode),
+                        QSSGRhiHelpers::toRhi(renderableImage->m_imageNode.m_verticalTilingMode),
+                        QSSGRhiHelpers::toRhi(renderableImage->m_imageNode.m_depthTilingMode)
                     };
                     rhiCtx->checkAndAdjustForNPoT(texture, &samplerDesc);
                     QRhiSampler *sampler = rhiCtx->sampler(samplerDesc);
@@ -465,41 +493,44 @@ void QSSGCustomMaterialSystem::rhiPrepareRenderable(QSSGRhiGraphicsPipelineState
             }
         }
 
+        QSSGRhiContextPrivate *rhiCtxD = QSSGRhiContextPrivate::get(rhiCtx);
+
         // do the same srb lookup acceleration as default materials
         QRhiShaderResourceBindings *&srb = dcd.srb;
         bool srbChanged = false;
         if (!srb || bindings != dcd.bindings) {
-            srb = rhiCtx->srb(bindings);
+            srb = rhiCtxD->srb(bindings);
+            rhiCtxD->releaseCachedSrb(dcd.bindings);
             dcd.bindings = bindings;
             srbChanged = true;
         }
 
-        if (cubeFace < 0)
+        if (cubeFace == QSSGRenderTextureCubeFaceNone)
             renderable.rhiRenderData.mainPass.srb = srb;
         else
-            renderable.rhiRenderData.reflectionPass.srb[cubeFace] = srb;
+            renderable.rhiRenderData.reflectionPass.srb[cubeFaceIdx] = srb;
 
-        const QSSGGraphicsPipelineStateKey pipelineKey = QSSGGraphicsPipelineStateKey::create(*ps, renderPassDescriptor, srb);
+        const auto pipelineKey = QSSGGraphicsPipelineStateKey::create(*ps, renderPassDescriptor, srb);
         if (dcd.pipeline
                 && !srbChanged
                 && dcd.renderTargetDescriptionHash == pipelineKey.extra.renderTargetDescriptionHash
                 && dcd.renderTargetDescription == pipelineKey.renderTargetDescription
                 && dcd.ps == *ps)
         {
-            if (cubeFace < 0)
+            if (cubeFace == QSSGRenderTextureCubeFaceNone)
                 renderable.rhiRenderData.mainPass.pipeline = dcd.pipeline;
             else
                 renderable.rhiRenderData.reflectionPass.pipeline = dcd.pipeline;
         } else {
-            if (cubeFace < 0) {
-                renderable.rhiRenderData.mainPass.pipeline = rhiCtx->pipeline(pipelineKey,
-                                                                              renderPassDescriptor,
-                                                                              srb);
+            if (cubeFace == QSSGRenderTextureCubeFaceNone) {
+                renderable.rhiRenderData.mainPass.pipeline = rhiCtxD->pipeline(pipelineKey,
+                                                                               renderPassDescriptor,
+                                                                               srb);
                 dcd.pipeline = renderable.rhiRenderData.mainPass.pipeline;
             } else {
-                renderable.rhiRenderData.reflectionPass.pipeline = rhiCtx->pipeline(pipelineKey,
-                                                                                    renderPassDescriptor,
-                                                                                    srb);
+                renderable.rhiRenderData.reflectionPass.pipeline = rhiCtxD->pipeline(pipelineKey,
+                                                                                     renderPassDescriptor,
+                                                                                     srb);
                 dcd.pipeline = renderable.rhiRenderData.reflectionPass.pipeline;
             }
 
@@ -514,12 +545,12 @@ void QSSGCustomMaterialSystem::setShaderResources(char *ubufData,
                                                   const QSSGRenderCustomMaterial &inMaterial,
                                                   const QByteArray &inPropertyName,
                                                   const QVariant &propertyValue,
-                                                  QSSGRenderShaderDataType inPropertyType,
+                                                  QSSGRenderShaderValue::Type inPropertyType,
                                                   QSSGRhiShaderPipeline &shaderPipeline)
 {
     Q_UNUSED(inMaterial);
 
-    if (inPropertyType == QSSGRenderShaderDataType::Texture) {
+    if (inPropertyType == QSSGRenderShaderValue::Texture) {
         QSSGRenderCustomMaterial::TextureProperty *textureProperty =
                 reinterpret_cast<QSSGRenderCustomMaterial::TextureProperty *>(propertyValue.value<void *>());
         QSSGRenderImage *image = textureProperty->texImage;
@@ -530,12 +561,12 @@ void QSSGCustomMaterialSystem::setShaderResources(char *ubufData,
                 const QSSGRhiTexture t = {
                     inPropertyName,
                     texture.m_texture,
-                    { toRhi(textureProperty->minFilterType),
-                      toRhi(textureProperty->magFilterType),
-                      textureProperty->mipFilterType != QSSGRenderTextureFilterOp::None ? toRhi(textureProperty->mipFilterType) : QRhiSampler::None,
-                      toRhi(textureProperty->horizontalClampType),
-                      toRhi(textureProperty->verticalClampType),
-                      QRhiSampler::Repeat
+                    { QSSGRhiHelpers::toRhi(textureProperty->minFilterType),
+                      QSSGRhiHelpers::toRhi(textureProperty->magFilterType),
+                      textureProperty->mipFilterType != QSSGRenderTextureFilterOp::None ? QSSGRhiHelpers::toRhi(textureProperty->mipFilterType) : QRhiSampler::None,
+                      QSSGRhiHelpers::toRhi(textureProperty->horizontalClampType),
+                      QSSGRhiHelpers::toRhi(textureProperty->verticalClampType),
+                      QSSGRhiHelpers::toRhi(textureProperty->zClampType)
                     }
                 };
                 shaderPipeline.addExtraTexture(t);
@@ -562,15 +593,16 @@ void QSSGCustomMaterialSystem::applyRhiShaderPropertyValues(char *ubufData,
 void QSSGCustomMaterialSystem::rhiRenderRenderable(QSSGRhiContext *rhiCtx,
                                                    QSSGSubsetRenderable &renderable,
                                                    bool *needsSetViewport,
-                                                   int cubeFace,
+                                                   QSSGRenderTextureCubeFace cubeFace,
                                                    const QSSGRhiGraphicsPipelineState &state)
 {
     QRhiGraphicsPipeline *ps = renderable.rhiRenderData.mainPass.pipeline;
     QRhiShaderResourceBindings *srb = renderable.rhiRenderData.mainPass.srb;
 
-    if (cubeFace >= 0) {
+    if (cubeFace != QSSGRenderTextureCubeFaceNone) {
+        const auto cubeFaceIdx = QSSGBaseTypeHelpers::indexOfCubeFace(cubeFace);
         ps = renderable.rhiRenderData.reflectionPass.pipeline;
-        srb = renderable.rhiRenderData.reflectionPass.srb[cubeFace];
+        srb = renderable.rhiRenderData.reflectionPass.srb[cubeFaceIdx];
     }
 
     if (!ps || !srb)
